"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const SchemeGenerator_1 = require("../../domain/schemeGeneration/SchemeGenerator");
const ClassScheme_1 = require("../../domain/schema/schema/ClassScheme");
const CustomType_1 = require("../../domain/schema/types/CustomType");
const VectorType_1 = require("../../domain/schema/types/VectorType");
const NumberType_1 = require("../../domain/schema/types/NumberType");
const StringType_1 = require("../../domain/schema/types/StringType");
const BooleanType_1 = require("../../domain/schema/types/BooleanType");
const AnyType_1 = require("../../domain/schema/types/AnyType");
const ClassField_1 = require("../../domain/schema/schema/ClassField");
const Utils_1 = require("../codeGeneration/Utils");
class JSONSchemeGenerator extends SchemeGenerator_1.SchemeGenerator {
    generate(source) {
        return this.json2scheme(source.jsonData);
    }
    json2scheme(json, name = 'Root', schemesByName = {}) {
        let schemes = [];
        let fields = [];
        for (let key in json) {
            let value = json[key];
            let type = this.genTypeForPrimitive(json[key], Utils_1.toCamelCase(key, true));
            let customType = this.hasCustomType(type, value);
            if (customType) {
                let customTypeSchemes;
                if (value instanceof Array) {
                    customTypeSchemes = this.json2scheme(customType.value, Utils_1.toCamelCase(key, true), schemesByName);
                }
                else
                    customTypeSchemes = this.json2scheme(value, Utils_1.toCamelCase(key, true), schemesByName);
                // get scheme for current key (its the last one)
                let scheme = customTypeSchemes[customTypeSchemes.length - 1];
                // remove last scheme
                customTypeSchemes = customTypeSchemes.slice(0, customTypeSchemes.length - 1);
                if (schemesByName[scheme.name]) {
                    if (!this.isClassSchemesEqual(schemesByName[scheme.name], scheme)) {
                        // if we had that name and schemes are not the same -> we change name
                        let newName = Utils_1.toCamelCase(key + this.getRandomInt(1, 1000), true);
                        customTypeSchemes.push(new ClassScheme_1.default(newName, scheme.fields));
                        type = this.genTypeForPrimitive(json[key], newName);
                    }
                }
                else
                    customTypeSchemes.push(scheme);
                schemesByName[scheme.name] = scheme;
                schemes.push(...customTypeSchemes);
            }
            fields.push(new ClassField_1.default(key, type, ''));
        }
        schemes.push(new ClassScheme_1.default(name, fields));
        return schemes;
    }
    genTypeForPrimitive(value, name) {
        let type = typeof value;
        if (type == 'string')
            return new StringType_1.default();
        if (type == 'number')
            return new NumberType_1.default();
        if (type == 'boolean')
            return new BooleanType_1.default();
        if (value == null || value == undefined)
            return new AnyType_1.default();
        if (value instanceof Array) {
            if (value.length == 0)
                return new VectorType_1.default(new AnyType_1.default());
            let isUniform = true;
            let prevType = null;
            let prevItem;
            for (let item of value) {
                let type = this.genTypeForPrimitive(item, name);
                if (prevType && !this.isTypeEqual(type, prevType))
                    isUniform = false;
                if (prevItem &&
                    type instanceof CustomType_1.default &&
                    !this.isObjectsEqual(prevItem, item))
                    isUniform = false;
                prevType = type;
                prevItem = item;
            }
            if (isUniform)
                return new VectorType_1.default(prevType);
            return new VectorType_1.default(new AnyType_1.default());
        }
        if (value instanceof Object) {
            return new CustomType_1.default(name);
        }
        throw new Error('Unknown type for value:' + value);
    }
    isTypeEqual(firstType, secondType) {
        if (firstType instanceof NumberType_1.default &&
            secondType instanceof NumberType_1.default)
            return true;
        if (firstType instanceof StringType_1.default &&
            secondType instanceof StringType_1.default)
            return true;
        if (firstType instanceof BooleanType_1.default &&
            secondType instanceof BooleanType_1.default)
            return true;
        if (firstType instanceof AnyType_1.default &&
            secondType instanceof AnyType_1.default)
            return true;
        if (firstType instanceof VectorType_1.default &&
            secondType instanceof VectorType_1.default) {
            return this.isTypeEqual(firstType.item, secondType.item);
        }
        if (firstType instanceof CustomType_1.default &&
            secondType instanceof CustomType_1.default)
            return firstType.name == secondType.name;
        return false;
    }
    isObjectsEqual(firstObject, secondObject) {
        let firstObjectKeys = Object.keys(firstObject);
        let secondObjectKeys = Object.keys(secondObject);
        if (firstObjectKeys.length != secondObjectKeys.length)
            return false;
        for (let key in firstObjectKeys) {
            let firstObjectKey = firstObjectKeys[key];
            let secondObjectKey = secondObjectKeys[key];
            if (firstObjectKey !== secondObjectKey)
                return false;
            let firstType = this.genTypeForPrimitive(firstObject[firstObjectKey], firstObjectKey);
            let secondType = this.genTypeForPrimitive(secondObject[secondObjectKey], secondObjectKey);
            if (!this.isTypeEqual(firstType, secondType))
                return false;
        }
        return true;
    }
    isClassSchemesEqual(firstClassScheme, secondClassScheme) {
        if (firstClassScheme.name !== secondClassScheme.name)
            return false;
        if (firstClassScheme.fields.length !== secondClassScheme.fields.length)
            return false;
        for (let i in firstClassScheme.fields) {
            if (firstClassScheme.fields[i].name !== secondClassScheme.fields[i].name)
                return false;
            if (!this.isTypeEqual(firstClassScheme.fields[i].type, secondClassScheme.fields[i].type))
                return false;
        }
        return true;
    }
    hasCustomType(type, value) {
        if (type instanceof VectorType_1.default)
            return this.hasCustomType(type.item, value[0]);
        if (type instanceof CustomType_1.default)
            return { type, value };
        return null;
    }
    getRandomInt(min, max) {
        return Math.floor(Math.random() * (max - min)) + min;
    }
}
exports.default = JSONSchemeGenerator;
//# sourceMappingURL=JSONSchemeGenerator.js.map